package pfc;
import java.util.Calendar;
import java.util.Vector;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlType;

import pfc.conf.Constants;
import pfc.exception.PeticioConsultaException;

/** pfc.BeanPeticioConsulta
 * 
 * Dades de la peticio de consulta (ciutats, dades, habitacions, etc)
 */
@XmlType(propOrder = {"dataInici","dataFi","nits","origen","desti","vDistribucioHabitacions","vTxtDistribucioHabitacions"})
public class BeanPeticioConsulta {

	//Identificador de la peticio.
	private long id = -1;

	private Calendar cDesde = null;
	private Calendar cHasta = null;

	@XmlElement
	private int nits = -1;

	@XmlElement
	private String origen = null;
	@XmlElement
	private String desti = null;

	//Distribucio d'habitacions. Cada posicio indica la distribucio d'una habitacio.
	@XmlElement(name="distribucio")
	private Vector <DistribucioHabitacio> vDistribucioHabitacions = null;
	
	//Vector amb una descripcio textual de la distribucio de cada habitacio
	//p.e, {"1 habitacio de 1 adult"}
	//Utilitzat en la creacio de l'XML i de l'HTML resultat
	@XmlElementWrapper(name ="txtDistribucions")
	@XmlElement(name="txtDistribucio")
	private Vector <String> vTxtDistribucioHabitacions = null;

	/**
	 * Constructor necessari per la generacio de l'XML
	 */
	public BeanPeticioConsulta(){}

	/**
	 * Constructor de la peticio.
	 * Rep tots els parametres que identifiquen la peticio.
	 */
	public BeanPeticioConsulta(
		int diaDesde, int mesDesde, int anyDesde, 
		int diaHasta, int mesHasta, int anyHasta,
		String origen, String desti,
		Vector <DistribucioHabitacio> vDistribucioHabitacions,
		Vector<String> vTxtDistribucioHabitacions
	)
	{
		//L'identificador s'associa amb el moment de creacio de la peticio
		this.id = Calendar.getInstance().getTimeInMillis();

		this.cDesde = Calendar.getInstance();
		this.cDesde.set(Calendar.DAY_OF_MONTH,	diaDesde);
		this.cDesde.set(Calendar.MONTH,			mesDesde);
		this.cDesde.set(Calendar.YEAR,			anyDesde);

		this.cHasta = Calendar.getInstance();
		this.cHasta.set(Calendar.DAY_OF_MONTH,	diaHasta);
		this.cHasta.set(Calendar.MONTH,			mesHasta);
		this.cHasta.set(Calendar.YEAR,			anyHasta);

		this.origen = origen;
		this.desti = desti;

		this.vDistribucioHabitacions = vDistribucioHabitacions;
		
		this.vTxtDistribucioHabitacions = vTxtDistribucioHabitacions;

		//El nombre de nits de l'estancia es la diferencia entre la data final i la inicial. 
		//La diferencia en milisegons es pasa a dies.
		this.nits = (int) ((cHasta.getTimeInMillis() - cDesde.getTimeInMillis()) / (86400000));
	}

	/** Verificacio de que els parametres de consulta son correctes
	 * @throws PeticioConsultaException Si la consulta no compleix les restriccions del cercador
	 */
	public void assertPeticio(int diaInici, int mesInici, int anyInici, int diaFi, int mesFi, int anyFi) throws PeticioConsultaException {

		//Validacio que les dates siguin correctes (que els dies existeixen dins el mes i el any seleccionats)
		//Per exemple, no s'admet el 30 de Febrer o el 31 de Juny
		if (cDesde.get(Calendar.DAY_OF_MONTH)!=diaInici || cDesde.get(Calendar.MONTH)!=mesInici || cDesde.get(Calendar.YEAR)!=anyInici) {
			throw new PeticioConsultaException(Constants.MSG_ERROR_DATA_INICI_INCORRECTA);
		}
		if (cHasta.get(Calendar.DAY_OF_MONTH)!=diaFi || cHasta.get(Calendar.MONTH)!=mesFi || cHasta.get(Calendar.YEAR)!=anyFi) {
			throw new PeticioConsultaException(Constants.MSG_ERROR_DATA_FI_INCORRECTA);
		}
		
		//No es pot reservar per a dates passades
		int nitsFinsDesde = (int) ((cDesde.getTimeInMillis() - Calendar.getInstance().getTimeInMillis()) / (86400000));
		if (nitsFinsDesde <= 0) throw new PeticioConsultaException(Constants.MSG_ERROR_DATES_ANTERIORS_AVUI);

		//La data inicial ha de ser posterior a la data final
		if (nits <= 0) throw new PeticioConsultaException(Constants.MSG_ERROR_NITSNEGATIU);

		//S'imposa un nombre maxim de nits per estancia
		if (nits > Constants.MAX_NITS) throw new PeticioConsultaException(Constants.MSG_ERROR_NITEXCEDIDES);

		//No es pot reservar amb mes d'un any i mig d'antelacio
		Calendar limitDataInici = Calendar.getInstance();
		limitDataInici.add(Calendar.YEAR, 1);
		limitDataInici.add(Calendar.MONTH, 6);
		if (this.getcDesde().after(limitDataInici)){
			throw new PeticioConsultaException(Constants.MSG_ERROR_DATAINICITARD);
		}

		//L'origen no pot ser igual que el desti
		if (origen.equals(desti)) throw new PeticioConsultaException(Constants.MSG_ERROR_CIUTATSIGUALS);

		/*
		Els aeroports d'origen i desti no poden ser els mateixos
		Davant la situacio actual de configuracio de ciutats aixo no pot ocorrer
		No obstant, amb configuracions futures es podria donar el cas de viatjar entre ciutats
		diferents a les quals els correspon el mateix aeroport
		*/
		if (pfc.conf.Ciutats.getAeroport(this.origen).equals(pfc.conf.Ciutats.getAeroport(this.desti)))
			throw new PeticioConsultaException(Constants.MSG_ERROR_AEROPORTSIGUALS);

		//No s'ha indicat cap habitacio
		if (this.vDistribucioHabitacions.size()<=0)
			throw new PeticioConsultaException(Constants.MSG_ERROR_SENSEHABITACIONS);

		//Restriccio de nombre maxim de passatgers
		//La majoria d'operadors de vols imposen aquesta restriccio
		if (this.getMaxPax()>Constants.MAX_PAX)
			throw new PeticioConsultaException(Constants.MSG_ERROR_PAXEXCEDITS);

		//No es permeten habitacions sense adults
		if (this.getMinAdults()<=0)
			throw new PeticioConsultaException(Constants.MSG_ERROR_MINADULTS);
	}

	/**
	 * @return Nombre d'habitacions de la peticio
	 */
	public int getNumHabitacions() {
		return (this.vDistribucioHabitacions==null) ? 0 : this.vDistribucioHabitacions.size() ;
	}

	/**
	 * @param numHab Index de l'habitacio
	 * @return Nombre d'adults a l'habitacio indicada
	 */
	public int getNumAdultsHabitacio(int numHab){
		return this.vDistribucioHabitacions.get(numHab).getNumAdults();
	}

	/**
	 * @param numHab Index de l'habitacio
	 * @param edat Edat minima d'un adult
	 * @return Nombre d'adults a l'habitacio indicada
	 */
	public int getNumAdultsHabitacio(int numHab, int edat){
		return this.vDistribucioHabitacions.get(numHab).getNumAdults(edat);
	}

	/**
	 * @param numHab Index de l'habitacio
	 * @return Nombre de nins a l'habitacio indicada
	 */
	public int getNumNinsHabitacio(int numHab){
		return this.vDistribucioHabitacions.get(numHab).getNumNins();
	}

	/**
	 * @param numHab Index de l'habitacio
	 * @param edat Edat minima d'un adult
	 * @return Nombre de nins a l'habitacio indicada
	 */
	public int getNumNinsHabitacio(int numHab, int edat){
		return this.vDistribucioHabitacions.get(numHab).getNumNins(edat);
	}

	/**
	 * @param numHab Index de l'habitacio
	 * @param nin Index del nin
	 * @return Edat del nin indicat a l'habitacio indicada
	 */
	public int getEdatNin(int numHab, int nin){
		return this.vDistribucioHabitacions.get(numHab).getEdatNin(nin);
	}
	
	/**
	 * @param numHab Index de l'habitacio
	 * @param nin Index del nin
	 * @param edat Edat minima d'un nin
	 * @return Edat del nin indicat a l'habitacio indicada
	 */
	public int getEdatNin(int numHab, int nin, int edat){
		return this.vDistribucioHabitacions.get(numHab).getEdatNin(nin,edat);
	}

	/**
	 * @param edat Edat minima d'un adult
	 * @return Maxim nombre d'adults a qualsevol de les habitacions
	 */
	public int getMaxAdults(int edat) {
		int numAdults = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			if (distribucio.getNumAdults(edat)>numAdults)
				numAdults = distribucio.getNumAdults(edat);
		}
		return numAdults;
	}

	/**
	 * @return Maxim nombre d'adults a qualsevol de les habitacions
	 */
	public int getMaxAdults() {
		return getMaxAdults(pfc.conf.Constants.DEFAULT_EDAT_ADULT);
	}

	/**
	 * @param edat Edat minima d'un nin
	 * @return Maxim nombre de nins a qualsevol de les habitacions
	 */
	public int getMaxNins(int edat) {
		int numNins = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			if (distribucio.getNumNins(edat)>numNins)
				numNins = distribucio.getNumNins(edat);
		}
		return numNins;
	}

	/**
	 * @return Maxim nombre de nins a qualsevol de les habitacions
	 */
	public int getMaxNins() {
		return getMaxNins(pfc.conf.Constants.DEFAULT_EDAT_ADULT);
	}

	/**
	 * @param edatAdult Edat minima d'un adult
	 * @return Minim nombre d'adults a qualsevol de les habitacions
	 */
	private int getMinAdults(int edatAdult) {
		int numAdults = -1;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			if (numAdults==-1 || distribucio.getNumAdults(edatAdult)<numAdults)
				numAdults = distribucio.getNumAdults(edatAdult);
		}
		return numAdults;
	}

	/**
	 * @return Minim nombre de nins a qualsevol de les habitacions
	 */
	private int getMinAdults() {
		return getMinAdults(pfc.conf.Constants.DEFAULT_EDAT_ADULT);
	}

	/**
	 * @param edat Edat minima d'un adult 
	 * @return Nombre total d'adults a la peticio
	 */
	public int getNumTotalAdults(int edat) {
		int result = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			result += distribucio.getNumAdults(edat);
		}	
		return result;
	}

	/**
	 * @param edat Edat miima d'un nin 
	 * @return Nombre total de nins a la peticio
	 */
	public int getNumTotalNins(int edat) {
		int result = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			result += distribucio.getNumNins(edat);
		}	
		return result;
	}
	
	/**
	 * @param edatInicial Edat minima de l'ocupant
	 * @param edatFinal Edat maxima de l'ocupant
	 * @return Nombre de persones a la peticio amb edats segons els parametres
	 */
	public int getNumTotalPax(int edatInicial, int edatFinal) {
		int result = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			result += distribucio.getNumTotalPax(edatInicial,edatFinal);
		}	
		return result;
	}

	/**
	 * @return Nombre de nins a la peticio
	 */
	public int getNumTotalNins() {
		return getNumTotalNins(pfc.conf.Constants.DEFAULT_EDAT_ADULT);
	}

	/**
	 * @param edat Edat minima d'un infant
	 * @return Nombre d'infants a la peticio
	 */
	public int getNumTotalBebes(int edat) {
		int result = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions)
		{
			result += distribucio.getNumBebes(edat);
		}	
		return result;
	}
	
	/**
	 * @return Nombre total d'ocupants/passatgers a la peticio
	 */
	public int getNumTotalPax(){
		int result = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions) {
			result += distribucio.numPax();
		}	
		return result;
	}
	
	/**
	 * @return Maxima ocupacio a qualque habitacio
	 */
	private int getMaxPax(){
		int result = 0;
		for (DistribucioHabitacio distribucio: this.vDistribucioHabitacions) {
			if (distribucio.numPax()>result) result = distribucio.numPax();
		}	
		return result;
	}	

	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append("\n[");
		sb.append("\n\tcid                        = " + this.id);
		sb.append("\n\tcDesde                     = " + pfc.conf.Constants.sDf.format(this.cDesde.getTime()));
		sb.append("\n\tcHasta                     = " + pfc.conf.Constants.sDf.format(this.cHasta.getTime()));
		sb.append("\n\tnits                       = " + this.getNits());
		sb.append("\n\torigen                     = " + this.origen);
		sb.append("\n\tdesti                      = " + this.desti);		
		sb.append("\n\tnumHabitacions             = " + this.getNumHabitacions());
		sb.append("\n\tnumTotalPax                = " + this.getNumTotalPax());
		sb.append("\n\tmaxAdults/Hab              = " + this.getMaxAdults());
		sb.append("\n\tmaxNins/Hab                = " + this.getMaxNins());
		sb.append("\n\tminAdults/Hab              = " + this.getMinAdults());
		sb.append("\n\tgetNumTotalAdults(>18)     = " + this.getNumTotalAdults(18));
		sb.append("\n\tnumTotalPax(2,17)          = " + this.getNumTotalPax(2,17));
		sb.append("\n\tnumTotalBebes(1)           = " + this.getNumTotalBebes(1));
		sb.append("\n\tmaxPax()                   = " + this.getMaxPax());
		sb.append("\n\tvDistribucioHabitacions    = " + vDistribucioHabitacions);
		sb.append("\n\tvTxtDistribucioHabitacions = " + vTxtDistribucioHabitacions);
		sb.append("\n]");
		return sb.toString();
	}

	/**
	 * @return Ciutat d'origen
	 */
	public String getOrigen() {
		return origen;
	}

	/**
	 * @return Ciutat de desti
	 */
	public String getDesti() {
		return desti;
	}

	/**
	 * @return Nombre de nits de la peticio
	 */
	public int getNits() {
		return nits;
	}

	/**
	 * @return Identificador de la peticio
	 */
	public long getId() {
		return id;
	}

	/**
	 * @return Data inicial
	 */
	public Calendar getcDesde() {
		return cDesde;
	}

	/**
	 * @return Data final
	 */
	public Calendar getcHasta() {
		return cHasta;
	}

	@XmlElement (name = "data_inicial")
	public String getDataInici(){
		return pfc.conf.Constants.sDf.format(this.cDesde.getTime());
	}
	
	@XmlElement (name = "data_fi")
	public String getDataFi(){
		return pfc.conf.Constants.sDf.format(this.cHasta.getTime());
	}

}