package pfc;

import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.client.cache.CacheConfig;
import org.apache.http.impl.client.cache.CachingHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import pfc.bean.CombinacioBean;
import pfc.bean.ItemBean;
import pfc.bean.ResultatBean;
import pfc.exception.SenseResultatsException;
import pfc.http.GestorConnexionsExpirades;
import pfc.http.HttpLog;
import pfc.util.FileUtil;
import pfc.wrapper.Wrapper;

/** pfc.Cercador
 * 
 * Implementa el cercador. 
 * Te la responsabilitat d'inicialitzar les dades generals del cercador i de fer la cerca.
 * Durant la cerca, la seva responsabilitat ee crear els diferents wrappers, recollir els resultats de la cerca, fer les ordenacions i muntar l'XML i l'HTML de presentaci.
 */
public class Cercador {

	protected final static Logger loger = Logger.getLogger("searcher");

	//Unica instancia del cercador
	private static Cercador cercador = null;

	//Vectors amb els noms de les classes dels operadors actius
	//Cada thread de cerca instancia els seus objectes dels wrappers d'operadors
	private static Vector<String> vWrappersHotels = null;
	private static Vector<String> vWrappersVols = null;

	//Vectors amb l'html del scraping dels operadors (si el debug esta activat)
	//A cada posicio del vector hi ha el contingut html d'un fitxer diferent
	private static Vector<String> scrapingHtmlHotels = null;
	private static Vector<String> scrapingHtmlVols = null;

	//Variables globals de gestio de les connexions http
	private static DefaultHttpClient httpclient = null;
	private static CachingHttpClient cachingHttpClient = null;
	private static GestorConnexionsExpirades gestorConnexionsExpirades = null;

	//Cache de l'XSLT
	private static Templates templatesXSLT = null;

	/** Retorna l'unica instancia del cercador
	 * @return Instancia del cercador
	 * @throws Exception
	 */
	public synchronized static Cercador getInstance() throws Exception {
		if (cercador == null){
			cercador = new Cercador();
		}
		return cercador;
	}

	/** Libera recursos i tanca connexions
	 * @throws Exception
	 */
	public synchronized static void shutdown() {
		if (cercador!=null) {
			try {
				//Tancament de totes les connexions i alliberament de recursos
			     httpclient.getConnectionManager().shutdown();
			     gestorConnexionsExpirades.shutdown();
			     //S'espera a la finalitzacio del thread
			     gestorConnexionsExpirades.join();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
			finally {
				cercador = null;
			}
		}
		System.out.println("Cercador.shutdown() cercador is down");
	}

	private Cercador() throws Exception {

		System.out.println("Cercador.Cercador() INIT");

		//Aquest tipus de manager s'usa quan mes d'un thread accedeix a l'objecte HttpClient
        ThreadSafeClientConnManager cm = new ThreadSafeClientConnManager();
        cm.setMaxTotal(100);
        CacheConfig cacheConfig = new CacheConfig();
        cacheConfig.setMaxCacheEntries(pfc.conf.Constants.CACHE_MAX_ENTRIES);
        cacheConfig.setMaxObjectSizeBytes(pfc.conf.Constants.CACHE_MAX_OBJECTSIZE);
        gestorConnexionsExpirades = new GestorConnexionsExpirades(cm);

        //Inicialitzacio dels objectes principals emprata a les connexions
	   System.out.println("Cercador.Cercador() init httpclient");
        httpclient = new DefaultHttpClient(cm);
        cachingHttpClient = new CachingHttpClient(httpclient, cacheConfig);
        cachingHttpClient.getParams().setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 4000);
        //cachingHttpClient.getParams().setParameter(CoreConnectionPNames.SO_TIMEOUT, 5000);

        //Carrega del fitxer de configuracio (fitxer de propietats)
        System.out.println("Cercador.Cercador() init config.properties");
        InputStream inputStreamCONFIG = Thread.currentThread().getContextClassLoader().getResourceAsStream(pfc.conf.Constants.PATH_PROPERTIES_CONFIG);
        Properties perfilBundle = new Properties();
        perfilBundle.load(inputStreamCONFIG);

        System.out.println("Cercador.Cercador() init properties");

        //Path XML
        FileUtil.setPathXML(perfilBundle.getProperty("PATH_XML"));

		//Path debug
		FileUtil.setPathDebug(perfilBundle.getProperty("PATH_DEBUG"));

		//Es comprova si el contingut dels preus es llegueix de fitxers que contenen codi html
		try {
			String debugScrapingActivat = perfilBundle.getProperty("debugScrapingActivat");
			HttpLog.setDebugScrapingActivat(Boolean.parseBoolean(debugScrapingActivat));
			System.out.println("debugScrapingActivat: '"+HttpLog.isDebugScrapingActivat()+"'");
		} catch (Exception e) {
			HttpLog.setDebugScrapingActivat(false);
			System.out.println("debugScrapingActivat property not found");
		}

		//Es comprova si el debug d'html de les peticions http esta activat
		try {
			String debugHtmlHTTPActivat = perfilBundle.getProperty("debugHtmlHTTPActivat");
			HttpLog.setDebugHtmlHttpActivat(Boolean.parseBoolean(debugHtmlHTTPActivat));
			System.out.println("debugHtmlHTTPActivat: '"+HttpLog.isDebugHtmlHTTPActivat()+"'");
		} catch (Exception e) {
			System.out.println("debugHtmlHTTPActivat property not found");
		}

		//Es comprova si el debug d'html de les operacions amb expressions regulars esta activat
		try {
			String debugHtmlItemsActivat = perfilBundle.getProperty("debugHtmlItemsActivat");
			HttpLog.setDebugHtmlItemsActivat(Boolean.parseBoolean(debugHtmlItemsActivat));
			System.out.println("debugHtmlItemsActivat: '"+HttpLog.isDebugHtmlItemsActivat()+"'");
		} catch (Exception e) {
			System.out.println("debugHtmlItemsActivat property not found");
		}

		//Es comprova si el debug de cookies esta activat
		try {
			String debugCookiesActivat = perfilBundle.getProperty("debugCookiesActivat");
			HttpLog.setDebugCookiesActivat(Boolean.parseBoolean(debugCookiesActivat));
			System.out.println("debugCookiesActivat: '"+HttpLog.isDebugCookiesActivat()+"'");
		} catch (Exception e) {
			System.out.println("debugCookiesActivat property not found");
		}

		//Lectura dels wrappers d'hotels actius
		vWrappersHotels = new Vector<String>();
		scrapingHtmlHotels = new Vector<String>();
		StringTokenizer stHotels = new StringTokenizer(perfilBundle.getProperty("wrappersHotelsActius"),";");
		while (stHotels.hasMoreTokens()) {
			String nextToken = "pfc.wrapper.hotel.hotels." + stHotels.nextToken();
			System.out.println("ACTIVAT "+nextToken);
			vWrappersHotels.add(nextToken);
			if (HttpLog.isDebugScrapingActivat()) {
				scrapingHtmlHotels.add(FileUtil.readDebugHTML(perfilBundle.getProperty("scrapingFile_" + nextToken)));
			}
		}

		//Lectura dels wrappers de vols actius
		vWrappersVols = new Vector<String>();
		scrapingHtmlVols = new Vector<String>();
		StringTokenizer stVols = new StringTokenizer(perfilBundle.getProperty("wrappersVolsActius"),";");
		while (stVols.hasMoreTokens()) {
			String nextToken = "pfc.wrapper.flight.flights." + stVols.nextToken();
			System.out.println("ACTIVAT "+nextToken);
			vWrappersVols.add(nextToken);
			if (HttpLog.isDebugScrapingActivat()) {
				scrapingHtmlVols.add(FileUtil.readDebugHTML(perfilBundle.getProperty("scrapingFile_" + nextToken)));
			}
		}

		//Lectura i analisi de la plantilla XSLT
		System.out.println("Cercador.Cercador() Analitzant plantilla XSLT...");
		InputStream inputStreamXSLT = Thread.currentThread().getContextClassLoader().getResourceAsStream(pfc.conf.Constants.PATH_XSLT);
		templatesXSLT = TransformerFactory.newInstance().newTemplates(new StreamSource(inputStreamXSLT));

		//Inici del thread gestorConnexionsExpirades
		System.out.println("Cercador.Cercador() Iniciant proces periodic...");
		gestorConnexionsExpirades.start();
	}

	public String search(BeanPeticioConsulta beanPeticioConsulta) throws SenseResultatsException, Exception {
		try 
		{
			loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] search(peticio="+beanPeticioConsulta+")");
			System.out.println("["+beanPeticioConsulta.getId()+"] launching threads(peticio="+beanPeticioConsulta+")");

			//S'inicien els threads que cerquen estancies a hotels
			loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] launching hotels threads...");
			Vector<Wrapper> vWrappersHotelsPeticio = new Vector<Wrapper>();
			for (int i=0; i<vWrappersHotels.size(); i++) {
				Wrapper wrapper = (Wrapper)Class.forName(vWrappersHotels.get(i)).newInstance();
				wrapper.setBeanPeticioConsulta(beanPeticioConsulta);
				wrapper.setHttpClient(cachingHttpClient);
				if (HttpLog.isDebugScrapingActivat()) {
					wrapper.setHtmlDebugScraping(scrapingHtmlHotels.get(i));
				}

				vWrappersHotelsPeticio.add(wrapper);
				System.out.println("["+beanPeticioConsulta.getId()+"] launching hotel thread (" + wrapper.getClass().getName() + ")...");
				//It is never legal to start a thread more than once. In particular, a thread may not be restarted once it has completed execution.
				wrapper.start(); //IllegalThreadStateException
				Thread.sleep(pfc.conf.Constants.TIME_BETWEEN_STARTS);			
			}

			//S'inicien els threads que cerquen vols
			loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] launching flight threads...");
			Vector<Wrapper> vWrappersVolsPeticio = new Vector<Wrapper>();
			for (int i=0; i<vWrappersVols.size(); i++) {
				Wrapper wrapper = (Wrapper)Class.forName(vWrappersVols.get(i)).newInstance();
				wrapper.setBeanPeticioConsulta(beanPeticioConsulta);
				wrapper.setHttpClient(cachingHttpClient);
				if (HttpLog.isDebugScrapingActivat()) {
					wrapper.setHtmlDebugScraping(scrapingHtmlVols.get(i));
				}

				vWrappersVolsPeticio.add(wrapper);
				System.out.println("["+beanPeticioConsulta.getId()+"] launching flight thread (" + wrapper.getClass().getName() + ")...");
				wrapper.start();
				Thread.sleep(pfc.conf.Constants.TIME_BETWEEN_STARTS);
			}

			//S'espera a la finalitzacio de l'execucio de tots els threads
			System.out.println("["+beanPeticioConsulta.getId()+"] Waiting threads to end...");
			loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] Waiting for hotels threads...");
			for (Wrapper wrapper: vWrappersHotelsPeticio) {
				wrapper.join();
			}
			loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] Waiting for flight threads...");
			for (Wrapper wrapper: vWrappersVolsPeticio) {
				wrapper.join();
			}

			HashMap<String,ItemBean> hmHotelBean = new HashMap<String,ItemBean>();
			Vector<ItemBean> vDebugHotelBean = new Vector<ItemBean>();
			for (Wrapper wrapper: vWrappersHotelsPeticio) {
				Vector<ItemBean> vHotelBean = wrapper.getResultatScraping();
				vDebugHotelBean.addAll(vHotelBean);
				FileUtil.printHTMLItems(vHotelBean,wrapper.getClass().getName()+"#items.html");
				loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] ["+wrapper.getClass().getName() + ":" + vHotelBean.size() + " hotels found]");
				for (ItemBean itemBean: vHotelBean){
					ItemBean elementActual = hmHotelBean.get(itemBean.getKey());
					if (elementActual!=null){
						loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] Overlapped hotels ["+elementActual+":"+itemBean+"]["+itemBean.getKey()+"]");
						if (elementActual.compareTo(itemBean)==1){
							hmHotelBean.put(itemBean.getKey().toUpperCase(), itemBean);
						}
					}
					else hmHotelBean.put(itemBean.getKey(), itemBean);
				}
			}

			Vector<ItemBean> vDebugFlightBean = new Vector<ItemBean>();
			for (Wrapper wrapper: vWrappersVolsPeticio) {
				Vector<ItemBean> vVolBean = wrapper.getResultatScraping();
				vDebugFlightBean.addAll(vVolBean);
				FileUtil.printHTMLItems(vVolBean,wrapper.getClass().getName()+"#items.html");
				loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] ["+wrapper.getClass().getName() + ":" + vVolBean.size() + " flights found]");
			}

			Vector<ItemBean> vFinalHotelBean = new Vector<ItemBean>(hmHotelBean.values());
			Vector<ItemBean> vFinalFlightBean = new Vector<ItemBean>(vDebugFlightBean);			

			//No es fan les combinacions si no es troben hotels o vols
			if (vFinalHotelBean.size() == 0 && vFinalFlightBean.size() == 0) {
				loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] NO RESULTS FOUND");
				throw new SenseResultatsException(pfc.conf.Constants.MSG_ERROR_SENSE_RESULTATS);
			}
			else {
				if (vFinalHotelBean.size() == 0) {
					loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] NO HOTEL FOUND");
					throw new SenseResultatsException(pfc.conf.Constants.MSG_ERROR_SENSE_HOTELS);
				}
				if (vFinalFlightBean.size() == 0) {
					loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] NO FLIGHTS FOUND");
					throw new SenseResultatsException(pfc.conf.Constants.MSG_ERROR_SENSE_VOLS);
				}
			}

			System.out.println("["+beanPeticioConsulta.getId()+"] ordering cheapest prices...");

			//Ordenacio hotels
			loger.log(Level.DEBUG, "Ordenacio hotels... ["+vFinalHotelBean.size() + " hotels trobats]");
			FileUtil.printHTMLItems(vDebugHotelBean,"AllHotelItems.html");
			Collections.sort(vFinalHotelBean);
			FileUtil.printHTMLItems(vFinalHotelBean,"AllHotelItems#sorted.html");

			//Ordenacio vols			
			loger.log(Level.DEBUG, "Ordenacio vols... ["+vFinalFlightBean.size() + " vols trobats]");
			FileUtil.printHTMLItems(vDebugFlightBean,"AllFlightItems.html");
			Collections.sort(vFinalFlightBean);
			FileUtil.printHTMLItems(vFinalFlightBean,"AllFlightItems#sorted.html");

			//Combinacio hotels
			vFinalHotelBean  = new Vector<ItemBean>(vFinalHotelBean.subList(0, (vFinalHotelBean.size()<5?vFinalHotelBean.size():5)));
			FileUtil.printHTMLItems(vFinalHotelBean,"AllHotelItems#sorted#final.html");

			//Combinacio vols
			vFinalFlightBean = new Vector<ItemBean>(vFinalFlightBean.subList(0, (vFinalFlightBean.size()<5?vFinalFlightBean.size():5)));
			FileUtil.printHTMLItems(vFinalFlightBean,"AllFlightItems#sorted#final.html");

			//Combinacio hotels i vols
			loger.log(Level.DEBUG, "["+beanPeticioConsulta.getId()+"] Looking for final cheapest combinations...");
			Vector<CombinacioBean> vCombinacions = new Vector<CombinacioBean>();
			for (ItemBean hotelBean: vFinalHotelBean){
				for (ItemBean volBean: vFinalFlightBean){
					vCombinacions.add(new CombinacioBean(hotelBean, volBean));
				}
			}
			FileUtil.printHTMLItems(vCombinacions,"combinations.html");

			Collections.sort(vCombinacions);
			loger.log(Level.DEBUG, "Nombre combinacions hotel+vol trobades: '" + vCombinacions.size() + "'");
			FileUtil.printHTMLItems(vCombinacions,"combinations#sorted.html");

			//Resultat final
			vCombinacions = new Vector<CombinacioBean>(vCombinacions.subList(0, (vCombinacions.size()<5?vCombinacions.size():5)));
			FileUtil.printHTMLItems(vCombinacions,"combinations#sorted#final.html");

			System.out.println("["+beanPeticioConsulta.getId()+"] generating XML...");
			String xml = generarXML(beanPeticioConsulta, vCombinacions);
			FileUtil.printXML(xml,beanPeticioConsulta.getId()+".xml");

			String html = generarHTML(xml);
			FileUtil.printHTML(html,"result.html");

			return html;
		}
		catch (Exception e){
			throw e;
		}
	}

	/**
	 * @param beanPeticioConsulta
	 * @param vCombinacions
	 * @return
	 * @throws JAXBException Si es produeix qualque error en les crides a l'API JAXB
	 * @throws FileNotFoundException Error en la creacio del fitxer
	 * @throws UnsupportedEncodingException Si la codificacio de l'arxiu creat no existeix o no esta disponible
	 */
	private String generarXML(BeanPeticioConsulta beanPeticioConsulta, Vector<CombinacioBean> vCombinacions) throws Exception
	{
	    ResultatBean resultatBean = new ResultatBean();
	    resultatBean.setBeanPeticioConsulta(beanPeticioConsulta);
	    resultatBean.setvCombinacions(vCombinacions);

	    JAXBContext contextObj = JAXBContext.newInstance(pfc.bean.ResultatBean.class,ItemBean.class,pfc.bean.hotel.HotelBean.class,pfc.bean.flight.FlightBean.class);
	    Marshaller marshallerObj = contextObj.createMarshaller();
	    marshallerObj.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
	    StringWriter oWriter = new StringWriter();
	    marshallerObj.marshal(resultatBean, oWriter);

	    String xml = oWriter.toString();

		return xml;
	}

	private String generarHTML(String xml) throws Exception 
	{
		StringReader oReader = new StringReader(xml);
		StreamSource streamSource = new StreamSource(oReader);

		StringWriter oWriter = new StringWriter();
		StreamResult streamResult = new StreamResult(oWriter);

		Transformer transformer = templatesXSLT.newTransformer();
		transformer.transform(streamSource, streamResult);

		return oWriter.toString();
	}

}